(* generated by Ott 0.21 from: lang.ott *)

Require Import Arith.
Require Import Bool.
Require Import List.


Require Import Omega ClassicalChoice basis. 
Definition termvar := nat.
Lemma eq_termvar: forall (x y : termvar), {x = y} + {x <> y}.
Proof.
  decide equality; auto with ott_coq_equality arith.
Defined.
Hint Resolve eq_termvar : ott_coq_equality.
Definition typename := nat.
Lemma eq_typename: forall (x y : typename), {x = y} + {x <> y}.
Proof.
  decide equality; auto with ott_coq_equality arith.
Defined.
Hint Resolve eq_typename : ott_coq_equality.
Definition typebvar := nat.
Lemma eq_typebvar: forall (x y : typebvar), {x = y} + {x <> y}.
Proof.
  decide equality; auto with ott_coq_equality arith.
Defined.
Hint Resolve eq_typebvar : ott_coq_equality.
Definition location := nat.

Inductive exp : Set :=  (*r expressions *)
 | Evar (x:termvar) (*r Variable *)
 | Eunit : exp
 | Eint (n:nat)
 | Etrue : exp
 | Efalse : exp
 | Epair (e1:exp) (e2:exp)
 | Einl (e:exp)
 | Einr (e:exp)
 | Efst (e:exp)
 | Esnd (e:exp)
 | Eloc (loc:location)
 | Eif (e0:exp) (e1:exp) (e2:exp) (*r if *)
 | Ecase (x:termvar) (e0:exp) (e1:exp) (e2:exp) (*r case *)
 | Eroll (e:exp) (*r roll *)
 | Eunroll (e:exp) (*r unroll *)
 | Efix (f:termvar) (x:termvar) (e:exp) (*r Recursive function *)
 | Eapp (e:exp) (e':exp) (*r Function application *)
 | Egen (e:exp) (*r Type generalisation *)
 | Einst (e:exp) (*r Type instantiation *)
 | Epack (e:exp)
 | Eunpack (x:termvar) (e1:exp) (e2:exp)
 | Enew (e:exp)
 | Eread (e:exp)
 | Ewrite (e1:exp) (e2:exp)
 | Erefeq (e1:exp) (e2:exp).

Inductive typ : Set :=  (*r types (in locally nameless style) *)
 | Tname (aa:typename) (*r Free name *)
 | Tbvar (bb:typebvar) (*r Bound variable *)
 | Tunit : typ (*r Unit *)
 | Tint : typ (*r Integers *)
 | Tbool : typ (*r Booleans *)
 | Tprod (ty1:typ) (ty2:typ) (*r products *)
 | Tsum (ty1:typ) (ty2:typ) (*r Sums *)
 | Tarrow (ty1:typ) (ty2:typ) (*r Functions *)
 | Tref (ty:typ) (*r References *)
 | Tmu (ty:typ) (*r Recursive types *)
 | Tforall (ty:typ)
 | Texists (ty:typ).

Inductive val : Set :=  (*r values *)
 | Vvar (x:termvar)
 | Vunit : val
 | Vint (n:nat)
 | Vtrue : val
 | Vfalse : val
 | Vpair (v1:val) (v2:val)
 | Vinl (v:val)
 | Vinr (v:val)
 | Vroll (v:val)
 | Vfix (f:termvar) (x:termvar) (e:exp)
 | Vgen (e:exp)
 | Vpack (v:val)
 | Vloc (loc:location).

Inductive prg : Set :=  (*r Programs *)
 | Pvar (x:termvar)
 | Punit : prg
 | Pint (n:nat)
 | Ptrue : prg
 | Pfalse : prg
 | Ppair (p1:prg) (p2:prg)
 | Pinl (ty:typ) (p:prg)
 | Pinr (ty:typ) (p:prg)
 | Pfst (p:prg)
 | Psnd (p:prg)
 | Pif (p0:prg) (p1:prg) (p2:prg)
 | Pcase (x:termvar) (p0:prg) (p1:prg) (p2:prg)
 | Proll (ty:typ) (p:prg)
 | Punroll (p:prg)
 | Pfix (f:termvar) (x:termvar) (ty:typ) (ty':typ) (p:prg)
 | Papp (p:prg) (p':prg) (*r Function application *)
 | Pgen (p:prg)
 | Pinst (p:prg) (ty:typ)
 | Ppack (ty1:typ) (p:prg) (ty:typ)
 | Punpack (x:termvar) (p1:prg) (p2:prg)
 | Pnew (p:prg)
 | Pread (p:prg)
 | Pwrite (p1:prg) (p2:prg)
 | Prefeq (p1:prg) (p2:prg).

Inductive cont : Set :=  (*r Multi-step evaluation contexts *)
 | Khole : cont
 | Kif (K:cont) (e1:exp) (e2:exp)
 | KpairL (K:cont) (e:exp)
 | KpairR (v:val) (K:cont)
 | Kfst (K:cont)
 | Ksnd (K:cont)
 | Kinl (K:cont)
 | Kinr (K:cont)
 | Kcase (x:termvar) (K:cont) (e1:exp) (e2:exp)
 | Kroll (K:cont)
 | Kunroll (K:cont)
 | Kpack (K:cont)
 | KappL (K:cont) (e:exp)
 | KappR (v:val) (K:cont)
 | Kinst (K:cont)
 | Kunpack (x:termvar) (K:cont) (e:exp)
 | Knew (K:cont)
 | Kread (K:cont)
 | KwriteL (K:cont) (e:exp)
 | KwriteR (v:val) (K:cont)
 | KrefeqL (K:cont) (e:exp)
 | KrefeqR (v:val) (K:cont).

Inductive scont : Set :=  (*r Single-step evaluation contexts *)
 | KKif (e1:exp) (e2:exp)
 | KKpairL (e:exp)
 | KKpairR (v:val)
 | KKfst : scont
 | KKsnd : scont
 | KKinl : scont
 | KKinr : scont
 | KKcase (x:termvar) (e1:exp) (e2:exp)
 | KKroll : scont
 | KKunroll : scont
 | KKpack : scont
 | KKappL (e:exp)
 | KKappR (v:val)
 | KKinst : scont
 | KKunpack (x:termvar) (e:exp)
 | KKnew : scont
 | KKread : scont
 | KKwriteL (e:exp)
 | KKwriteR (v:val)
 | KKrefeqL (e:exp)
 | KKrefeqR (v:val).

Inductive ctx : Set :=  (*r Contexts *)
 | Chole : ctx
 | Cpair1 (C:ctx) (p2:prg)
 | Cpair2 (p1:prg) (C:ctx)
 | Cinl (ty:typ) (C:ctx)
 | Cinr (ty:typ) (C:ctx)
 | Cfst (C:ctx)
 | Csnd (C:ctx)
 | Cif0 (C:ctx) (p1:prg) (p2:prg)
 | Cif1 (p0:prg) (C:ctx) (p2:prg)
 | Cif2 (p0:prg) (p1:prg) (C:ctx)
 | Ccase0 (x:termvar) (C:ctx) (p1:prg) (p2:prg)
 | Ccase1 (x:termvar) (p0:prg) (C:ctx) (p2:prg)
 | Ccase2 (x:termvar) (p0:prg) (p1:prg) (C:ctx)
 | Croll (ty:typ) (C:ctx)
 | Cunroll (C:ctx)
 | Cfix (f:termvar) (x:termvar) (ty:typ) (ty':typ) (C:ctx)
 | Capp1 (C:ctx) (p:prg)
 | Capp2 (p:prg) (C:ctx)
 | Cgen (C:ctx)
 | Cinst (C:ctx) (ty:typ)
 | Cpack (ty1:typ) (C:ctx) (ty:typ)
 | Cunpack1 (x:termvar) (C:ctx) (p:prg)
 | Cunpack2 (x:termvar) (p:prg) (C:ctx)
 | Cnew (C:ctx)
 | Cread (C:ctx)
 | Cwrite1 (C:ctx) (p:prg)
 | Cwrite2 (p:prg) (C:ctx)
 | Crefeq1 (C:ctx) (p:prg)
 | Crefeq2 (p:prg) (C:ctx).

Definition typeenv : Set := (list typename).

Definition termenv : Set := (list (termvar * typ)).


(*----------------------------------------------------------------------------*)
(** Definition of a ``locally closed'' type *)
(*----------------------------------------------------------------------------*)

Fixpoint lc_typ (n : nat) (ty : typ) :=
  match ty with
  | Tbvar aa     => aa < n
  | Tname aa    => True
  | Tunit       => True 
  | Tint        => True 
  | Tbool       => True
  | (Tprod ty1 ty2)  => lc_typ n ty1 /\ lc_typ n ty2
  | (Tsum ty1 ty2)   => lc_typ n ty1 /\ lc_typ n ty2
  | (Tarrow ty1 ty2) => lc_typ n ty1 /\ lc_typ n ty2
  | (Tref ty)    => lc_typ n ty
  | (Tmu ty)     => lc_typ (S n) ty
  | (Tforall ty) => lc_typ (S n) ty
  | (Texists ty) => lc_typ (S n) ty
  end.

(*----------------------------------------------------------------------------*)
(** Substitution of bounded type variables *)
(*----------------------------------------------------------------------------*)

Fixpoint vsubst (x: typebvar) (tyA ty: typ) :=
  match ty with
  | Tbvar aa      =>
    match nat_compare aa x with
      | Lt => Tbvar aa
      | Eq => tyA
      | Gt => Tbvar (pred aa)
    end
  | Tname aa    => Tname aa
  | Tunit       => Tunit 
  | Tint        => Tint 
  | Tbool       => Tbool 
  | (Tprod ty1 ty2)  => Tprod  (vsubst x tyA ty1) (vsubst x tyA ty2)
  | (Tsum ty1 ty2)   => Tsum   (vsubst x tyA ty1) (vsubst x tyA ty2)
  | (Tarrow ty1 ty2) => Tarrow (vsubst x tyA ty1) (vsubst x tyA ty2)
  | (Tref ty)    => Tref (vsubst x tyA ty)
  | (Tmu ty)     => Tmu     (vsubst (S x) tyA ty)
  | (Tforall ty) => Tforall (vsubst (S x) tyA ty)
  | (Texists ty) => Texists (vsubst (S x) tyA ty)
  end.

(*----------------------------------------------------------------------------*)
(** Definition of heaps *)
(*----------------------------------------------------------------------------*)

Definition finite_domain A B (f : A -> option B) := 
  exists d, forall x y, f x = Some y -> In x d.
Implicit Arguments finite_domain.

Structure heap := 
  { heap_f :> location -> option val 
  ; heap_findom : finite_domain heap_f }.

Program Definition hemp : heap := {| heap_f := fun l => None |}.
Next Obligation. exists nil; discriminate. Qed.

Program Definition hext (h : heap) (loc : location) (v: val) : heap :=
  {| heap_f := fun l => if eq_nat_dec l loc then Some v else h l |}.  
Next Obligation. 
  destruct (heap_findom h) as [d X]; exists (loc :: d); intros; simpl.
  destruct (eq_nat_dec); eauto.
Qed.

Definition hplus (h1 h2 h: heap) := 
  forall x, h x = h2 x /\ h1 x = None \/ h x = h1 x /\ h2 x = None.


Parameter allocator : heap -> location.

Axiom allocator_unalloc : forall h : heap, h (allocator h) = None.

Lemma allocator_consistency: 
  exists allocator : heap -> location, forall h : heap, h (allocator h) = None.
Proof.
  cut (forall h : heap, exists l, h l = None).
    intros FRESH; apply (choice _ FRESH).
  intros [hf [d FIN] ]; simpl.
  set (l := S (fold_right plus 0 d)); specialize (FIN l); exists l.
  destruct (hf l); auto.
  cut (forall n (GE: n >= l) (IN: In n d), False).
    intros; exfalso; eauto.
  unfold l in *; clear -d; induction d; intros; eauto.
  simpl in GE, IN; destruct IN; [subst n | apply (IHd n)]; auto; omega.
Qed.

(*----------------------------------------------------------------------------*)
(** Coercions *)
(*----------------------------------------------------------------------------*)

Fixpoint exp_of_val (v : val) : exp :=
  match v with
  | Vvar x      => Evar x
  | Vunit       => Eunit
  | Vint n      => Eint n
  | Vtrue       => Etrue 
  | Vfalse      => Efalse
  | Vpair v1 v2 => Epair (exp_of_val v1) (exp_of_val v2)
  | Vinl v      => Einl (exp_of_val v)
  | Vinr v      => Einr (exp_of_val v)
  | Vroll v     => Eroll (exp_of_val v)
  | Vfix f x e  => Efix f x e
  | Vgen e      => Egen e
  | Vpack v     => Epack (exp_of_val v)
  | Vloc loc    => Eloc loc
  end.

Coercion exp_of_val : val >-> exp.

Fixpoint app_cont (K : cont) (e0 : exp) : exp :=
  match K with
  | Khole => e0
  | (Kif K e1 e2) => (Eif  (  (app_cont  K   e0 )  )  e1 e2)
  | (KpairL K e) => (Epair  (  (app_cont  K   e0 )  )  e)
  | (KpairR v K) => (Epair  (exp_of_val  v )   (  (app_cont  K   e0 )  ) )
  | (Kfst K) => (Efst  (  (app_cont  K   e0 )  ) )
  | (Ksnd K) => (Esnd  (  (app_cont  K   e0 )  ) )
  | (Kinl K) => (Einl  (  (app_cont  K   e0 )  ) )
  | (Kinr K) => (Einr  (  (app_cont  K   e0 )  ) )
  | (Kcase x K e1 e2) => (Ecase x  (  (app_cont  K   e0 )  )  e1 e2)
  | (Kroll K) => (Eroll  (  (app_cont  K   e0 )  ) )
  | (Kunroll K) => (Eunroll  (  (app_cont  K   e0 )  ) )
  | (Kpack K) => (Epack  (  (app_cont  K   e0 )  ) )
  | (KappL K e) => (Eapp  (  (app_cont  K   e0 )  )  e)
  | (KappR v K) => (Eapp  (exp_of_val  v )   (  (app_cont  K   e0 )  ) )
  | (Kinst K) => (Einst  (  (app_cont  K   e0 )  ) )
  | (Kunpack x K e) => (Eunpack x  (  (app_cont  K   e0 )  )  e)
  | (Knew K) => (Enew  (  (app_cont  K   e0 )  ) )
  | (Kread K) => (Eread  (  (app_cont  K   e0 )  ) )
  | (KwriteL K e) => (Ewrite  (  (app_cont  K   e0 )  )  e)
  | (KwriteR v K) => (Ewrite  (exp_of_val  v )   (  (app_cont  K   e0 )  ) )
  | (KrefeqL K e) => (Erefeq  (  (app_cont  K   e0 )  )  e)
  | (KrefeqR v K) => (Erefeq  (exp_of_val  v )   (  (app_cont  K   e0 )  ) )
  end.

Fixpoint list_scont_of_cont (K : cont) : list scont :=
  match K with
  | Khole => nil
  | (Kif K e1 e2) => cons (KKif e1 e2) (list_scont_of_cont K)
  | (KpairL K e) => cons (KKpairL e) (list_scont_of_cont K)
  | (KpairR v K) => cons (KKpairR v) (list_scont_of_cont K)
  | (Kfst K) => cons KKfst (list_scont_of_cont K)
  | (Ksnd K) => cons KKsnd (list_scont_of_cont K)
  | (Kinl K) => cons KKinl (list_scont_of_cont K)
  | (Kinr K) => cons KKinr (list_scont_of_cont K)
  | (Kcase x K e1 e2) => cons (KKcase x e1 e2) (list_scont_of_cont K)
  | (Kroll K) => cons KKroll (list_scont_of_cont K)
  | (Kunroll K) => cons KKunroll (list_scont_of_cont K)
  | (Kpack K) => cons KKpack (list_scont_of_cont K)
  | (KappL K e) => cons (KKappL e) (list_scont_of_cont K)
  | (KappR v K) => cons (KKappR v) (list_scont_of_cont K)
  | (Kinst K) => cons KKinst (list_scont_of_cont K)
  | (Kunpack x K e) => cons (KKunpack x e) (list_scont_of_cont K)
  | (Knew K) => cons KKnew (list_scont_of_cont K)
  | (Kread K) => cons KKread (list_scont_of_cont K)
  | (KwriteL K e) => cons (KKwriteL e) (list_scont_of_cont K)
  | (KwriteR v K) => cons (KKwriteR v) (list_scont_of_cont K)
  | (KrefeqL K e) => cons (KKrefeqL e) (list_scont_of_cont K)
  | (KrefeqR v K) => cons (KKrefeqR v) (list_scont_of_cont K)
  end.

Fixpoint cont_of_list_scont (KK : list scont) : cont :=
  match KK with
  | nil => Khole
  | cons (KKif e1 e2) KK => let K := cont_of_list_scont KK in (Kif K e1 e2)
  | cons (KKpairL e) KK => let K := cont_of_list_scont KK in (KpairL K e)
  | cons (KKpairR v) KK => let K := cont_of_list_scont KK in (KpairR v K)
  | cons KKfst KK => let K := cont_of_list_scont KK in (Kfst K)
  | cons KKsnd KK => let K := cont_of_list_scont KK in (Ksnd K)
  | cons KKinl KK => let K := cont_of_list_scont KK in (Kinl K)
  | cons KKinr KK => let K := cont_of_list_scont KK in (Kinr K)
  | cons (KKcase x e1 e2) KK => let K := cont_of_list_scont KK in (Kcase x K e1 e2)
  | cons KKroll KK => let K := cont_of_list_scont KK in (Kroll K)
  | cons KKunroll KK => let K := cont_of_list_scont KK in (Kunroll K)
  | cons KKpack KK => let K := cont_of_list_scont KK in (Kpack K)
  | cons (KKappL e) KK => let K := cont_of_list_scont KK in (KappL K e)
  | cons (KKappR v) KK => let K := cont_of_list_scont KK in (KappR v K)
  | cons KKinst KK => let K := cont_of_list_scont KK in (Kinst K)
  | cons (KKunpack x e) KK => let K := cont_of_list_scont KK in (Kunpack x K e)
  | cons KKnew KK => let K := cont_of_list_scont KK in (Knew K)
  | cons KKread KK => let K := cont_of_list_scont KK in (Kread K)
  | cons (KKwriteL e) KK => let K := cont_of_list_scont KK in (KwriteL K e)
  | cons (KKwriteR v) KK => let K := cont_of_list_scont KK in (KwriteR v K)
  | cons (KKrefeqL e) KK => let K := cont_of_list_scont KK in (KrefeqL K e)
  | cons (KKrefeqR v) KK => let K := cont_of_list_scont KK in (KrefeqR v K)
  end.

Lemma cont_of_list_scont_of_cont :
  forall x, cont_of_list_scont (list_scont_of_cont x) = x.
Proof. induction x; simpl; f_equal; trivial. Qed.

Lemma list_cont_of_scont_of_list_cont :
  forall x, list_scont_of_cont (cont_of_list_scont x) = x.
Proof. induction x as [|[]?]; simpl; f_equal; trivial. Qed.

Fixpoint app_ctx (C : ctx) (p0: prg) : prg :=
  match C with
  | Chole => p0                         
  | (Cpair1 C p2) => (Ppair  (  (app_ctx  C   p0 )  )  p2)
  | (Cpair2 p1 C) => (Ppair p1  (  (app_ctx  C   p0 )  ) )
  | (Cinl ty C) => (Pinl ty  (  (app_ctx  C   p0 )  ) )
  | (Cinr ty C) => (Pinr ty  (  (app_ctx  C   p0 )  ) )
  | (Cfst C) => (Pfst  (  (app_ctx  C   p0 )  ) )
  | (Csnd C) => (Psnd  (  (app_ctx  C   p0 )  ) )
  | (Cif0 C p1 p2) => (Pif  (  (app_ctx  C   p0 )  )  p1 p2)
  | (Cif1 p C p2) => (Pif p  (  (app_ctx  C   p0 )  )  p2)
  | (Cif2 p p1 C) => (Pif p p1  (  (app_ctx  C   p0 )  ) )
  | (Ccase0 x C p1 p2) => (Pcase x  (  (app_ctx  C   p0 )  )  p1 p2)
  | (Ccase1 x p C p2) => (Pcase x p  (app_ctx  C   p0 )  p2)
  | (Ccase2 x p p1 C) => (Pcase x p p1  (app_ctx  C   p0 ) )
  | (Croll ty C) => (Proll ty  (  (app_ctx  C   p0 )  ) )
  | (Cunroll C) => (Punroll  (  (app_ctx  C   p0 )  ) )
  | (Cfix f x ty ty' C) => (Pfix f x ty ty'  (  (app_ctx  C   p0 )  ) )
  | (Capp1 C p) => (Papp  (  (app_ctx  C   p0 )  )  p)
  | (Capp2 p C) => (Papp p  (  (app_ctx  C   p0 )  ) )
  | (Cgen C) => (Pgen  (  (app_ctx  C   p0 )  ) )
  | (Cinst C ty) => (Pinst  (  (app_ctx  C   p0 )  )  ty)
  | (Cpack ty1 C ty) => (Ppack ty1  (app_ctx  C   p0 )  ty)
  | (Cunpack1 x C p) => (Punpack x  (app_ctx  C   p0 )  p)
  | (Cunpack2 x p C) => (Punpack x p  (  (app_ctx  C   p0 )  ) )
  | (Cnew C) => (Pnew  (  (app_ctx  C   p0 )  ) )
  | (Cread C) => (Pread  (  (app_ctx  C   p0 )  ) )
  | (Cwrite1 C p) => (Pwrite  (  (app_ctx  C   p0 )  )  p)
  | (Cwrite2 p C) => (Pwrite p  (  (app_ctx  C   p0 )  ) )
  | (Crefeq1 C p) => (Prefeq  (  (app_ctx  C   p0 )  )  p)
  | (Crefeq2 p C) => (Prefeq p  (  (app_ctx  C   p0 )  ) )
  end.

Fixpoint erase_prg (p : prg) : exp :=
  match p with
  | (Pvar x) => Evar x
  | Punit => Eunit
  | (Pint n) => Eint n
  | Ptrue => Etrue
  | Pfalse => Efalse
  | (Ppair p1 p2) => (Epair  (erase_prg  p1 )   (erase_prg  p2 ) )
  | (Pinl ty p) => (Einl  (erase_prg  p ) )
  | (Pinr ty p) => (Einr  (erase_prg  p ) )
  | (Pfst p) => (Efst  (erase_prg  p ) )
  | (Psnd p) => (Esnd  (erase_prg  p ) )
  | (Pif p0 p1 p2) => (Eif  (erase_prg  p0 )   (erase_prg  p1 )   (erase_prg  p2 ) )
  | (Pcase x p0 p1 p2) => (Ecase x  (erase_prg  p0 )   (erase_prg  p1 )   (erase_prg  p2 ) )
  | (Proll ty p) => (Eroll  (erase_prg  p ) )
  | (Punroll p) => (Eunroll  (erase_prg  p ) )
  | (Pfix f x ty ty' p) => (Efix f x  (erase_prg  p ) )
  | (Papp p p') => (Eapp  (erase_prg  p )   (erase_prg  p' ) )
  | (Pgen p) => (Egen  (erase_prg  p ) )
  | (Pinst p ty) => (Einst  (erase_prg  p ) )
  | (Ppack ty1 p ty) => (Epack  (erase_prg  p ) )
  | (Punpack x p1 p2) => (Eunpack x  (erase_prg  p1 )   (erase_prg  p2 ) )
  | (Pnew p) => (Enew  (erase_prg  p ) )
  | (Pread p) => (Eread  (erase_prg  p ) )
  | (Pwrite p1 p2) => (Ewrite  (erase_prg  p1 )   (erase_prg  p2 ) )
  | (Prefeq p1 p2) => (Erefeq  (erase_prg  p1 )   (erase_prg  p2 ) )
  end.

Coercion erase_prg : prg >-> exp.
 


(** subrules *)
Fixpoint is_value_of_exp (e_5:exp) : Prop :=
  match e_5 with
  | (Evar x) => (True)
  | Eunit => (True)
  | (Eint n) => (True)
  | Etrue => (True)
  | Efalse => (True)
  | (Epair e1 e2) => ((is_value_of_exp e1) /\ (is_value_of_exp e2))
  | (Einl e) => ((is_value_of_exp e))
  | (Einr e) => ((is_value_of_exp e))
  | (Efst e) => False
  | (Esnd e) => False
  | (Eloc loc) => (True)
  | (Eif e0 e1 e2) => False
  | (Ecase x e0 e1 e2) => False
  | (Eroll e) => ((is_value_of_exp e))
  | (Eunroll e) => False
  | (Efix f x e) => (True)
  | (Eapp e e') => False
  | (Egen e) => (True)
  | (Einst e) => False
  | (Epack e) => ((is_value_of_exp e))
  | (Eunpack x e1 e2) => False
  | (Enew e) => False
  | (Eread e) => False
  | (Ewrite e1 e2) => False
  | (Erefeq e1 e2) => False
end.


(** context application *)
Definition appctx_scont_exp (scont5:scont) (exp5:exp) : exp :=
  match scont5 with
  | (KKif e1 e2) => (Eif exp5 e1 e2)
  | (KKpairL e) => (Epair exp5 e)
  | (KKpairR v) => (Epair  (exp_of_val  v )  exp5)
  | KKfst => (Efst exp5)
  | KKsnd => (Esnd exp5)
  | KKinl => (Einl exp5)
  | KKinr => (Einr exp5)
  | (KKcase x e1 e2) => (Ecase x exp5 e1 e2)
  | KKroll => (Eroll exp5)
  | KKunroll => (Eunroll exp5)
  | KKpack => (Epack exp5)
  | (KKappL e) => (Eapp exp5 e)
  | (KKappR v) => (Eapp  (exp_of_val  v )  exp5)
  | KKinst => (Einst exp5)
  | (KKunpack x e) => (Eunpack x exp5 e)
  | KKnew => (Enew exp5)
  | KKread => (Eread exp5)
  | (KKwriteL e) => (Ewrite exp5 e)
  | (KKwriteR v) => (Ewrite  (exp_of_val  v )  exp5)
  | (KKrefeqL e) => (Erefeq exp5 e)
  | (KKrefeqR v) => (Erefeq  (exp_of_val  v )  exp5)
end.


(** free variables *)
Fixpoint ftn_typ (ty_5:typ) : list typename :=
  match ty_5 with
  | (Tname aa) => (cons aa nil)
  | (Tbvar bb) => nil
  | Tunit => nil
  | Tint => nil
  | Tbool => nil
  | (Tprod ty1 ty2) => (app (ftn_typ ty1) (ftn_typ ty2))
  | (Tsum ty1 ty2) => (app (ftn_typ ty1) (ftn_typ ty2))
  | (Tarrow ty1 ty2) => (app (ftn_typ ty1) (ftn_typ ty2))
  | (Tref ty) => ((ftn_typ ty))
  | (Tmu ty) => ((ftn_typ ty))
  | (Tforall ty) => ((ftn_typ ty))
  | (Texists ty) => ((ftn_typ ty))
end.

Fixpoint fv_exp (e_5:exp) : list termvar :=
  match e_5 with
  | (Evar x) => (cons x nil)
  | Eunit => nil
  | (Eint n) => nil
  | Etrue => nil
  | Efalse => nil
  | (Epair e1 e2) => (app (fv_exp e1) (fv_exp e2))
  | (Einl e) => ((fv_exp e))
  | (Einr e) => ((fv_exp e))
  | (Efst e) => ((fv_exp e))
  | (Esnd e) => ((fv_exp e))
  | (Eloc loc) => nil
  | (Eif e0 e1 e2) => (app (fv_exp e0) (app (fv_exp e1) (fv_exp e2)))
  | (Ecase x e0 e1 e2) => (app (fv_exp e0) (app (list_minus eq_termvar (fv_exp e1) (cons x nil)) (list_minus eq_termvar (fv_exp e2) (cons x nil))))
  | (Eroll e) => ((fv_exp e))
  | (Eunroll e) => ((fv_exp e))
  | (Efix f x e) => ((list_minus eq_termvar (fv_exp e) (app (cons x nil) (cons f nil))))
  | (Eapp e e') => (app (fv_exp e) (fv_exp e'))
  | (Egen e) => ((fv_exp e))
  | (Einst e) => ((fv_exp e))
  | (Epack e) => ((fv_exp e))
  | (Eunpack x e1 e2) => (app (fv_exp e1) (list_minus eq_termvar (fv_exp e2) (cons x nil)))
  | (Enew e) => ((fv_exp e))
  | (Eread e) => ((fv_exp e))
  | (Ewrite e1 e2) => (app (fv_exp e1) (fv_exp e2))
  | (Erefeq e1 e2) => (app (fv_exp e1) (fv_exp e2))
end.

Fixpoint ftn_prg (p_5:prg) : list typename :=
  match p_5 with
  | (Pvar x) => nil
  | Punit => nil
  | (Pint n) => nil
  | Ptrue => nil
  | Pfalse => nil
  | (Ppair p1 p2) => (app (ftn_prg p1) (ftn_prg p2))
  | (Pinl ty p) => (app (ftn_typ ty) (ftn_prg p))
  | (Pinr ty p) => (app (ftn_typ ty) (ftn_prg p))
  | (Pfst p) => ((ftn_prg p))
  | (Psnd p) => ((ftn_prg p))
  | (Pif p0 p1 p2) => (app (ftn_prg p0) (app (ftn_prg p1) (ftn_prg p2)))
  | (Pcase x p0 p1 p2) => (app (ftn_prg p0) (app (ftn_prg p1) (ftn_prg p2)))
  | (Proll ty p) => (app (ftn_typ ty) (ftn_prg p))
  | (Punroll p) => ((ftn_prg p))
  | (Pfix f x ty ty' p) => (app (ftn_typ ty) (app (ftn_typ ty') (ftn_prg p)))
  | (Papp p p') => (app (ftn_prg p) (ftn_prg p'))
  | (Pgen p) => ((ftn_prg p))
  | (Pinst p ty) => (app (ftn_prg p) (ftn_typ ty))
  | (Ppack ty1 p ty) => (app (ftn_typ ty1) (app (ftn_prg p) (ftn_typ ty)))
  | (Punpack x p1 p2) => (app (ftn_prg p1) (ftn_prg p2))
  | (Pnew p) => ((ftn_prg p))
  | (Pread p) => ((ftn_prg p))
  | (Pwrite p1 p2) => (app (ftn_prg p1) (ftn_prg p2))
  | (Prefeq p1 p2) => (app (ftn_prg p1) (ftn_prg p2))
end.

Fixpoint fv_val (v_5:val) : list termvar :=
  match v_5 with
  | (Vvar x) => nil
  | Vunit => nil
  | (Vint n) => nil
  | Vtrue => nil
  | Vfalse => nil
  | (Vpair v1 v2) => (app (fv_val v1) (fv_val v2))
  | (Vinl v) => ((fv_val v))
  | (Vinr v) => ((fv_val v))
  | (Vroll v) => ((fv_val v))
  | (Vfix f x e) => ((list_minus eq_termvar (fv_exp e) (app (cons x nil) (cons f nil))))
  | (Vgen e) => ((fv_exp e))
  | (Vpack v) => ((fv_val v))
  | (Vloc loc) => nil
end.

Fixpoint ftn_ctx (C5:ctx) : list typename :=
  match C5 with
  | Chole => nil
  | (Cpair1 C p2) => (app (ftn_ctx C) (ftn_prg p2))
  | (Cpair2 p1 C) => (app (ftn_prg p1) (ftn_ctx C))
  | (Cinl ty C) => (app (ftn_typ ty) (ftn_ctx C))
  | (Cinr ty C) => (app (ftn_typ ty) (ftn_ctx C))
  | (Cfst C) => ((ftn_ctx C))
  | (Csnd C) => ((ftn_ctx C))
  | (Cif0 C p1 p2) => (app (ftn_ctx C) (app (ftn_prg p1) (ftn_prg p2)))
  | (Cif1 p0 C p2) => (app (ftn_prg p0) (app (ftn_ctx C) (ftn_prg p2)))
  | (Cif2 p0 p1 C) => (app (ftn_prg p0) (app (ftn_prg p1) (ftn_ctx C)))
  | (Ccase0 x C p1 p2) => (app (ftn_ctx C) (app (ftn_prg p1) (ftn_prg p2)))
  | (Ccase1 x p0 C p2) => (app (ftn_prg p0) (app (ftn_ctx C) (ftn_prg p2)))
  | (Ccase2 x p0 p1 C) => (app (ftn_prg p0) (app (ftn_prg p1) (ftn_ctx C)))
  | (Croll ty C) => (app (ftn_typ ty) (ftn_ctx C))
  | (Cunroll C) => ((ftn_ctx C))
  | (Cfix f x ty ty' C) => (app (ftn_typ ty) (app (ftn_typ ty') (ftn_ctx C)))
  | (Capp1 C p) => (app (ftn_ctx C) (ftn_prg p))
  | (Capp2 p C) => (app (ftn_prg p) (ftn_ctx C))
  | (Cgen C) => ((ftn_ctx C))
  | (Cinst C ty) => (app (ftn_ctx C) (ftn_typ ty))
  | (Cpack ty1 C ty) => (app (ftn_typ ty1) (app (ftn_ctx C) (ftn_typ ty)))
  | (Cunpack1 x C p) => (app (ftn_ctx C) (ftn_prg p))
  | (Cunpack2 x p C) => (app (ftn_prg p) (ftn_ctx C))
  | (Cnew C) => ((ftn_ctx C))
  | (Cread C) => ((ftn_ctx C))
  | (Cwrite1 C p) => (app (ftn_ctx C) (ftn_prg p))
  | (Cwrite2 p C) => (app (ftn_prg p) (ftn_ctx C))
  | (Crefeq1 C p) => (app (ftn_ctx C) (ftn_prg p))
  | (Crefeq2 p C) => (app (ftn_prg p) (ftn_ctx C))
end.

Fixpoint fv_cont (K5:cont) : list termvar :=
  match K5 with
  | Khole => nil
  | (Kif K e1 e2) => (app (fv_cont K) (app (fv_exp e1) (fv_exp e2)))
  | (KpairL K e) => (app (fv_cont K) (fv_exp e))
  | (KpairR v K) => (app (fv_val v) (fv_cont K))
  | (Kfst K) => ((fv_cont K))
  | (Ksnd K) => ((fv_cont K))
  | (Kinl K) => ((fv_cont K))
  | (Kinr K) => ((fv_cont K))
  | (Kcase x K e1 e2) => (app (fv_cont K) (app (fv_exp e1) (fv_exp e2)))
  | (Kroll K) => ((fv_cont K))
  | (Kunroll K) => ((fv_cont K))
  | (Kpack K) => ((fv_cont K))
  | (KappL K e) => (app (fv_cont K) (fv_exp e))
  | (KappR v K) => (app (fv_val v) (fv_cont K))
  | (Kinst K) => ((fv_cont K))
  | (Kunpack x K e) => (app (fv_cont K) (fv_exp e))
  | (Knew K) => ((fv_cont K))
  | (Kread K) => ((fv_cont K))
  | (KwriteL K e) => (app (fv_cont K) (fv_exp e))
  | (KwriteR v K) => (app (fv_val v) (fv_cont K))
  | (KrefeqL K e) => (app (fv_cont K) (fv_exp e))
  | (KrefeqR v K) => (app (fv_val v) (fv_cont K))
end.

Definition fv_scont (KK5:scont) : list termvar :=
  match KK5 with
  | (KKif e1 e2) => (app (fv_exp e1) (fv_exp e2))
  | (KKpairL e) => ((fv_exp e))
  | (KKpairR v) => ((fv_val v))
  | KKfst => nil
  | KKsnd => nil
  | KKinl => nil
  | KKinr => nil
  | (KKcase x e1 e2) => (app (fv_exp e1) (fv_exp e2))
  | KKroll => nil
  | KKunroll => nil
  | KKpack => nil
  | (KKappL e) => ((fv_exp e))
  | (KKappR v) => ((fv_val v))
  | KKinst => nil
  | (KKunpack x e) => ((fv_exp e))
  | KKnew => nil
  | KKread => nil
  | (KKwriteL e) => ((fv_exp e))
  | (KKwriteR v) => ((fv_val v))
  | (KKrefeqL e) => ((fv_exp e))
  | (KKrefeqR v) => ((fv_val v))
end.


(** substitutions *)
Fixpoint subst_exp (sub:list (termvar*exp)) (e__6:exp) {struct e__6} : exp :=
  match e__6 with
  | (Evar x) => (match list_assoc eq_termvar x sub with None => (Evar x) | Some e_5 => e_5 end)
  | Eunit => Eunit 
  | (Eint n) => Eint n
  | Etrue => Etrue 
  | Efalse => Efalse 
  | (Epair e1 e2) => Epair (subst_exp sub e1) (subst_exp sub e2)
  | (Einl e) => Einl (subst_exp sub e)
  | (Einr e) => Einr (subst_exp sub e)
  | (Efst e) => Efst (subst_exp sub e)
  | (Esnd e) => Esnd (subst_exp sub e)
  | (Eloc loc) => Eloc loc
  | (Eif e0 e1 e2) => Eif (subst_exp sub e0) (subst_exp sub e1) (subst_exp sub e2)
  | (Ecase x e0 e1 e2) => Ecase x (subst_exp sub e0) (subst_exp (list_minus2 eq_termvar sub (cons x nil)) e1) (subst_exp (list_minus2 eq_termvar sub (cons x nil)) e2)
  | (Eroll e) => Eroll (subst_exp sub e)
  | (Eunroll e) => Eunroll (subst_exp sub e)
  | (Efix f x e) => Efix f x (subst_exp (list_minus2 eq_termvar sub (app (cons x nil) (cons f nil))) e)
  | (Eapp e e') => Eapp (subst_exp sub e) (subst_exp sub e')
  | (Egen e) => Egen (subst_exp sub e)
  | (Einst e) => Einst (subst_exp sub e)
  | (Epack e) => Epack (subst_exp sub e)
  | (Eunpack x e1 e2) => Eunpack x (subst_exp sub e1) (subst_exp (list_minus2 eq_termvar sub (cons x nil)) e2)
  | (Enew e) => Enew (subst_exp sub e)
  | (Eread e) => Eread (subst_exp sub e)
  | (Ewrite e1 e2) => Ewrite (subst_exp sub e1) (subst_exp sub e2)
  | (Erefeq e1 e2) => Erefeq (subst_exp sub e1) (subst_exp sub e2)
end.

Fixpoint tsubst_typ (sub:list (typename*typ)) (ty__6:typ) {struct ty__6} : typ :=
  match ty__6 with
  | (Tname aa) => (match list_assoc eq_typename aa sub with None => (Tname aa) | Some ty_5 => ty_5 end)
  | (Tbvar bb) => Tbvar bb
  | Tunit => Tunit 
  | Tint => Tint 
  | Tbool => Tbool 
  | (Tprod ty1 ty2) => Tprod (tsubst_typ sub ty1) (tsubst_typ sub ty2)
  | (Tsum ty1 ty2) => Tsum (tsubst_typ sub ty1) (tsubst_typ sub ty2)
  | (Tarrow ty1 ty2) => Tarrow (tsubst_typ sub ty1) (tsubst_typ sub ty2)
  | (Tref ty) => Tref (tsubst_typ sub ty)
  | (Tmu ty) => Tmu (tsubst_typ sub ty)
  | (Tforall ty) => Tforall (tsubst_typ sub ty)
  | (Texists ty) => Texists (tsubst_typ sub ty)
end.

Fixpoint subst_val (sub:list (termvar*exp)) (v_5:val) {struct v_5} : val :=
  match v_5 with
  | (Vvar x) => Vvar x
  | Vunit => Vunit 
  | (Vint n) => Vint n
  | Vtrue => Vtrue 
  | Vfalse => Vfalse 
  | (Vpair v1 v2) => Vpair (subst_val sub v1) (subst_val sub v2)
  | (Vinl v) => Vinl (subst_val sub v)
  | (Vinr v) => Vinr (subst_val sub v)
  | (Vroll v) => Vroll (subst_val sub v)
  | (Vfix f x e) => Vfix f x (subst_exp (list_minus2 eq_termvar sub (app (cons x nil) (cons f nil))) e)
  | (Vgen e) => Vgen (subst_exp sub e)
  | (Vpack v) => Vpack (subst_val sub v)
  | (Vloc loc) => Vloc loc
end.

Fixpoint tsubst_prg (sub:list (typename*typ)) (p_5:prg) {struct p_5} : prg :=
  match p_5 with
  | (Pvar x) => Pvar x
  | Punit => Punit 
  | (Pint n) => Pint n
  | Ptrue => Ptrue 
  | Pfalse => Pfalse 
  | (Ppair p1 p2) => Ppair (tsubst_prg sub p1) (tsubst_prg sub p2)
  | (Pinl ty p) => Pinl (tsubst_typ sub ty) (tsubst_prg sub p)
  | (Pinr ty p) => Pinr (tsubst_typ sub ty) (tsubst_prg sub p)
  | (Pfst p) => Pfst (tsubst_prg sub p)
  | (Psnd p) => Psnd (tsubst_prg sub p)
  | (Pif p0 p1 p2) => Pif (tsubst_prg sub p0) (tsubst_prg sub p1) (tsubst_prg sub p2)
  | (Pcase x p0 p1 p2) => Pcase x (tsubst_prg sub p0) (tsubst_prg sub p1) (tsubst_prg sub p2)
  | (Proll ty p) => Proll (tsubst_typ sub ty) (tsubst_prg sub p)
  | (Punroll p) => Punroll (tsubst_prg sub p)
  | (Pfix f x ty ty' p) => Pfix f x (tsubst_typ sub ty) (tsubst_typ sub ty') (tsubst_prg sub p)
  | (Papp p p') => Papp (tsubst_prg sub p) (tsubst_prg sub p')
  | (Pgen p) => Pgen (tsubst_prg sub p)
  | (Pinst p ty) => Pinst (tsubst_prg sub p) (tsubst_typ sub ty)
  | (Ppack ty1 p ty) => Ppack (tsubst_typ sub ty1) (tsubst_prg sub p) (tsubst_typ sub ty)
  | (Punpack x p1 p2) => Punpack x (tsubst_prg sub p1) (tsubst_prg sub p2)
  | (Pnew p) => Pnew (tsubst_prg sub p)
  | (Pread p) => Pread (tsubst_prg sub p)
  | (Pwrite p1 p2) => Pwrite (tsubst_prg sub p1) (tsubst_prg sub p2)
  | (Prefeq p1 p2) => Prefeq (tsubst_prg sub p1) (tsubst_prg sub p2)
end.

Fixpoint subst_cont (sub:list (termvar*exp)) (K5:cont) {struct K5} : cont :=
  match K5 with
  | Khole => Khole 
  | (Kif K e1 e2) => Kif (subst_cont sub K) (subst_exp sub e1) (subst_exp sub e2)
  | (KpairL K e) => KpairL (subst_cont sub K) (subst_exp sub e)
  | (KpairR v K) => KpairR (subst_val sub v) (subst_cont sub K)
  | (Kfst K) => Kfst (subst_cont sub K)
  | (Ksnd K) => Ksnd (subst_cont sub K)
  | (Kinl K) => Kinl (subst_cont sub K)
  | (Kinr K) => Kinr (subst_cont sub K)
  | (Kcase x K e1 e2) => Kcase x (subst_cont sub K) (subst_exp sub e1) (subst_exp sub e2)
  | (Kroll K) => Kroll (subst_cont sub K)
  | (Kunroll K) => Kunroll (subst_cont sub K)
  | (Kpack K) => Kpack (subst_cont sub K)
  | (KappL K e) => KappL (subst_cont sub K) (subst_exp sub e)
  | (KappR v K) => KappR (subst_val sub v) (subst_cont sub K)
  | (Kinst K) => Kinst (subst_cont sub K)
  | (Kunpack x K e) => Kunpack x (subst_cont sub K) (subst_exp sub e)
  | (Knew K) => Knew (subst_cont sub K)
  | (Kread K) => Kread (subst_cont sub K)
  | (KwriteL K e) => KwriteL (subst_cont sub K) (subst_exp sub e)
  | (KwriteR v K) => KwriteR (subst_val sub v) (subst_cont sub K)
  | (KrefeqL K e) => KrefeqL (subst_cont sub K) (subst_exp sub e)
  | (KrefeqR v K) => KrefeqR (subst_val sub v) (subst_cont sub K)
end.

Definition subst_scont (sub:list (termvar*exp)) (KK5:scont) : scont :=
  match KK5 with
  | (KKif e1 e2) => KKif (subst_exp sub e1) (subst_exp sub e2)
  | (KKpairL e) => KKpairL (subst_exp sub e)
  | (KKpairR v) => KKpairR (subst_val sub v)
  | KKfst => KKfst 
  | KKsnd => KKsnd 
  | KKinl => KKinl 
  | KKinr => KKinr 
  | (KKcase x e1 e2) => KKcase x (subst_exp sub e1) (subst_exp sub e2)
  | KKroll => KKroll 
  | KKunroll => KKunroll 
  | KKpack => KKpack 
  | (KKappL e) => KKappL (subst_exp sub e)
  | (KKappR v) => KKappR (subst_val sub v)
  | KKinst => KKinst 
  | (KKunpack x e) => KKunpack x (subst_exp sub e)
  | KKnew => KKnew 
  | KKread => KKread 
  | (KKwriteL e) => KKwriteL (subst_exp sub e)
  | (KKwriteR v) => KKwriteR (subst_val sub v)
  | (KKrefeqL e) => KKrefeqL (subst_exp sub e)
  | (KKrefeqR v) => KKrefeqR (subst_val sub v)
end.

Fixpoint tsubst_ctx (sub:list (typename*typ)) (C5:ctx) {struct C5} : ctx :=
  match C5 with
  | Chole => Chole 
  | (Cpair1 C p2) => Cpair1 (tsubst_ctx sub C) (tsubst_prg sub p2)
  | (Cpair2 p1 C) => Cpair2 (tsubst_prg sub p1) (tsubst_ctx sub C)
  | (Cinl ty C) => Cinl (tsubst_typ sub ty) (tsubst_ctx sub C)
  | (Cinr ty C) => Cinr (tsubst_typ sub ty) (tsubst_ctx sub C)
  | (Cfst C) => Cfst (tsubst_ctx sub C)
  | (Csnd C) => Csnd (tsubst_ctx sub C)
  | (Cif0 C p1 p2) => Cif0 (tsubst_ctx sub C) (tsubst_prg sub p1) (tsubst_prg sub p2)
  | (Cif1 p0 C p2) => Cif1 (tsubst_prg sub p0) (tsubst_ctx sub C) (tsubst_prg sub p2)
  | (Cif2 p0 p1 C) => Cif2 (tsubst_prg sub p0) (tsubst_prg sub p1) (tsubst_ctx sub C)
  | (Ccase0 x C p1 p2) => Ccase0 x (tsubst_ctx sub C) (tsubst_prg sub p1) (tsubst_prg sub p2)
  | (Ccase1 x p0 C p2) => Ccase1 x (tsubst_prg sub p0) (tsubst_ctx sub C) (tsubst_prg sub p2)
  | (Ccase2 x p0 p1 C) => Ccase2 x (tsubst_prg sub p0) (tsubst_prg sub p1) (tsubst_ctx sub C)
  | (Croll ty C) => Croll (tsubst_typ sub ty) (tsubst_ctx sub C)
  | (Cunroll C) => Cunroll (tsubst_ctx sub C)
  | (Cfix f x ty ty' C) => Cfix f x (tsubst_typ sub ty) (tsubst_typ sub ty') (tsubst_ctx sub C)
  | (Capp1 C p) => Capp1 (tsubst_ctx sub C) (tsubst_prg sub p)
  | (Capp2 p C) => Capp2 (tsubst_prg sub p) (tsubst_ctx sub C)
  | (Cgen C) => Cgen (tsubst_ctx sub C)
  | (Cinst C ty) => Cinst (tsubst_ctx sub C) (tsubst_typ sub ty)
  | (Cpack ty1 C ty) => Cpack (tsubst_typ sub ty1) (tsubst_ctx sub C) (tsubst_typ sub ty)
  | (Cunpack1 x C p) => Cunpack1 x (tsubst_ctx sub C) (tsubst_prg sub p)
  | (Cunpack2 x p C) => Cunpack2 x (tsubst_prg sub p) (tsubst_ctx sub C)
  | (Cnew C) => Cnew (tsubst_ctx sub C)
  | (Cread C) => Cread (tsubst_ctx sub C)
  | (Cwrite1 C p) => Cwrite1 (tsubst_ctx sub C) (tsubst_prg sub p)
  | (Cwrite2 p C) => Cwrite2 (tsubst_prg sub p) (tsubst_ctx sub C)
  | (Crefeq1 C p) => Crefeq1 (tsubst_ctx sub C) (tsubst_prg sub p)
  | (Crefeq2 p C) => Crefeq2 (tsubst_prg sub p) (tsubst_ctx sub C)
end.

Set Implicit Arguments. 
(** definitions *)

(* defns OpSemJudg *)
Inductive reduce : heap -> exp -> heap -> exp -> Prop :=    (* defn reduce *)
 | Rtrue : forall (h:heap) (e1 e2:exp),
     reduce h (Eif Etrue e1 e2) h e1
 | Rfalse : forall (h:heap) (e1 e2:exp),
     reduce h (Eif Efalse e1 e2) h e2
 | Rfst : forall (h:heap) (v1 v2:val),
     reduce h (Efst (Epair  (exp_of_val  v1 )   (exp_of_val  v2 ) )) h  (exp_of_val  v1 ) 
 | Rsnd : forall (h:heap) (v1 v2:val),
     reduce h (Esnd (Epair  (exp_of_val  v1 )   (exp_of_val  v2 ) )) h  (exp_of_val  v2 ) 
 | Rinl : forall (h:heap) (x:termvar) (v:val) (e1 e2:exp),
     reduce h (Ecase x (Einl  (exp_of_val  v ) ) e1 e2) h  (subst_exp(( x ,  (exp_of_val  v )  )::nil) e1 ) 
 | Rinr : forall (h:heap) (x:termvar) (v:val) (e1 e2:exp),
     reduce h (Ecase x (Einr  (exp_of_val  v ) ) e1 e2) h  (subst_exp(( x ,  (exp_of_val  v )  )::nil) e2 ) 
 | Rbeta : forall (h:heap) (f x:termvar) (e:exp) (v:val),
     reduce h (Eapp  ( (Efix f x e) )   (exp_of_val  v ) ) h  (subst_exp(( x ,  (exp_of_val  v )  )::( f , (Efix f x e) )::nil) e ) 
 | Rtybeta : forall (h:heap) (e:exp),
     reduce h (Einst  ( (Egen e) ) ) h e
 | Rpack : forall (h:heap) (x:termvar) (v:val) (e:exp),
     reduce h (Eunpack x (Epack  (exp_of_val  v ) ) e) h  (subst_exp(( x ,  (exp_of_val  v )  )::nil) e ) 
 | Rroll : forall (h:heap) (v:val),
     reduce h (Eunroll  ( (Eroll  (exp_of_val  v ) ) ) ) h  (exp_of_val  v ) 
 | Rnew : forall (h:heap) (v:val) (loc:location)
     (ALLOC:  (allocator  h  =  loc ) ),
     reduce h (Enew  (exp_of_val  v ) )  (hext  h   loc   v )  (Eloc loc)
 | Rread : forall (h:heap) (loc:location) (v:val)
     (LOAD:  ( h   loc  = Some  v ) ),
     reduce h (Eread (Eloc loc)) h  (exp_of_val  v ) 
 | Rwrite : forall (h:heap) (loc:location) (v v':val),
     reduce  (hext  h   loc   v )  (Ewrite (Eloc loc)  (exp_of_val  v' ) )  (hext  h   loc   v' )  Eunit
 | RrefeqT : forall (h:heap) (loc:location),
     reduce h (Erefeq (Eloc loc) (Eloc loc)) h Etrue
 | RrefeqF : forall (h:heap) (loc loc':location)
     (NEQ:  ( loc  <>  loc' ) ),
     reduce h (Erefeq (Eloc loc) (Eloc loc')) h Efalse
 | Rsctx : forall (h:heap) (KK:scont) (e:exp) (h':heap) (e':exp)
     (RED: reduce h e h' e'),
     reduce h  (appctx_scont_exp  KK   e )  h'  (appctx_scont_exp  KK   e' ) .
(** definitions *)

(* defns EnvOKjudg *)
Inductive wfenv : typeenv -> termenv -> Prop :=    (* defn wfenv *)
 | wfenv_ok : forall (D:typeenv) (G:termenv) (x:termvar) (ty:typ),
      (forall  x ,   (forall  ty ,   (  (In ( x , ty )  G )   ->   (lc_typ 0  ty )  )   )  )  ->
      (forall  x ,   (forall  ty ,   (  (In ( x , ty )  G )   ->   (list_minus eq_typename   (ftn_typ  ty )    D  = nil)  )   )  )  ->
     wfenv D G.
(** definitions *)

(* defns TypeOKjudg *)
Inductive wftyp : typeenv -> termenv -> typ -> Prop :=    (* defn wftyp *)
 | wftyp_ok : forall (D:typeenv) (G:termenv) (ty:typ),
     wfenv D G ->
      (lc_typ 0  ty )  ->
      (list_minus eq_typename   (ftn_typ  ty )    D  = nil)  ->
     wftyp D G ty.
(** definitions *)

(* defns TypingJudg *)
Inductive hastype : typeenv -> termenv -> exp -> typ -> Prop :=    (* defn hastype *)
 | HTvar : forall (D:typeenv) (G:termenv) (x:termvar) (ty:typ),
     wfenv D G ->
      (list_assoc eq_termvar  x   G  = Some  ty )  ->
     hastype D G (Evar x) ty
 | HTunit : forall (D:typeenv) (G:termenv),
     wfenv D G ->
     hastype D G Eunit Tunit
 | HTint : forall (D:typeenv) (G:termenv) (n:nat),
     wfenv D G ->
     hastype D G (Eint n) Tint
 | HTtrue : forall (D:typeenv) (G:termenv),
     wfenv D G ->
     hastype D G Etrue Tbool
 | HTfalse : forall (D:typeenv) (G:termenv),
     wfenv D G ->
     hastype D G Efalse Tbool
 | HTpair : forall (D:typeenv) (G:termenv) (e1 e2:exp) (ty1 ty2:typ),
     hastype D G e1 ty1 ->
     hastype D G e2 ty2 ->
     hastype D G (Epair e1 e2) (Tprod ty1 ty2)
 | HTfst : forall (D:typeenv) (G:termenv) (e:exp) (ty1 ty2:typ),
     hastype D G e (Tprod ty1 ty2) ->
     hastype D G (Efst e) ty1
 | HTsnd : forall (D:typeenv) (G:termenv) (e:exp) (ty2 ty1:typ),
     hastype D G e (Tprod ty1 ty2) ->
     hastype D G (Esnd e) ty2
 | HTinl : forall (D:typeenv) (G:termenv) (e:exp) (ty1 ty2:typ),
     hastype D G e ty1 ->
     wftyp D G ty2 ->
     hastype D G (Einl e) (Tsum ty1 ty2)
 | HTinr : forall (D:typeenv) (G:termenv) (e:exp) (ty1 ty2:typ),
     hastype D G e ty2 ->
     wftyp D G ty1 ->
     hastype D G (Einr e) (Tsum ty1 ty2)
 | HTif : forall (D:typeenv) (G:termenv) (e0 e1 e2:exp) (ty:typ),
     hastype D G e0 Tbool ->
     hastype D G e1 ty ->
     hastype D G e2 ty ->
     hastype D G (Eif e0 e1 e2) ty
 | HTcase : forall (D:typeenv) (G:termenv) (x:termvar) (e0 e1 e2:exp) (ty ty1 ty2:typ),
     hastype D G e0 (Tsum ty1 ty2) ->
     hastype D  (cons ( x , ty1 )  G )  e1 ty ->
     hastype D  (cons ( x , ty2 )  G )  e2 ty ->
     hastype D G (Ecase x e0 e1 e2) ty
 | HTroll : forall (D:typeenv) (G:termenv) (e:exp) (ty:typ),
     hastype D G e  (vsubst 0  (Tmu ty)   ty )  ->
     wftyp D G (Tmu ty) ->
     hastype D G (Eroll e) (Tmu ty)
 | HTunroll : forall (D:typeenv) (G:termenv) (e:exp) (ty:typ),
     hastype D G e (Tmu ty) ->
     hastype D G (Eunroll e)  (vsubst 0  (Tmu ty)   ty ) 
 | HTfix : forall (D:typeenv) (G:termenv) (f x:termvar) (e:exp) (ty1 ty2:typ),
     hastype D  (cons ( x , ty1 )   (cons ( f , (Tarrow ty1 ty2) )  G )  )  e ty2 ->
     wftyp D G ty1 ->
     wftyp D G ty2 ->
     hastype D G (Efix f x e) (Tarrow ty1 ty2)
 | HTapp : forall (D:typeenv) (G:termenv) (e e':exp) (ty2 ty1:typ),
     hastype D G e (Tarrow ty1 ty2) ->
     hastype D G e' ty1 ->
     hastype D G (Eapp e e') ty2
 | HTgen : forall (D:typeenv) (G:termenv) (e:exp) (ty:typ) (aa:typename),
     hastype  (cons  aa   D )  G e  (vsubst 0  (Tname aa)   ty )  ->
     wftyp D G (Tforall ty) ->
      (~ In  aa   D )  ->
     hastype D G (Egen e) (Tforall ty)
 | HTinst : forall (D:typeenv) (G:termenv) (e:exp) (ty' ty:typ),
     hastype D G e (Tforall ty) ->
     wftyp D G ty' ->
     hastype D G (Einst e)  (vsubst 0  ty'   ty ) 
 | HTpack : forall (D:typeenv) (G:termenv) (e:exp) (ty ty':typ),
     hastype D G e  (vsubst 0  ty'   ty )  ->
     wftyp D G ty' ->
     hastype D G (Epack e) (Texists ty)
 | HTunpack : forall (D:typeenv) (G:termenv) (x:termvar) (e1 e2:exp) (ty2 ty:typ) (aa:typename),
     hastype D G e1 (Texists ty) ->
     hastype  (cons  aa   D )   (cons ( x ,  (vsubst 0  (Tname aa)   ty )  )  G )  e2 ty2 ->
     wftyp D G ty2 ->
      (~ In  aa   D )  ->
     hastype D G (Eunpack x e1 e2) ty2
 | HTnew : forall (D:typeenv) (G:termenv) (e:exp) (ty:typ),
     hastype D G e ty ->
     hastype D G (Enew e) (Tref ty)
 | HTread : forall (D:typeenv) (G:termenv) (e:exp) (ty:typ),
     hastype D G e (Tref ty) ->
     hastype D G (Eread e) ty
 | HTwrite : forall (D:typeenv) (G:termenv) (e1 e2:exp) (ty:typ),
     hastype D G e1 (Tref ty) ->
     hastype D G e2 ty ->
     hastype D G (Ewrite e1 e2) Tunit
 | HTrefeq : forall (D:typeenv) (G:termenv) (e1 e2:exp) (ty:typ),
     hastype D G e1 (Tref ty) ->
     hastype D G e2 (Tref ty) ->
     hastype D G (Erefeq e1 e2) Tbool.
(** definitions *)

(* defns ProgTypingJudg *)
Inductive phastype : typeenv -> termenv -> prg -> typ -> Prop :=    (* defn phastype *)
 | PHTvar : forall (D:typeenv) (G:termenv) (x:termvar) (ty:typ),
     wfenv D G ->
      (list_assoc eq_termvar  x   G  = Some  ty )  ->
     phastype D G (Pvar x) ty
 | PHTunit : forall (D:typeenv) (G:termenv),
     wfenv D G ->
     phastype D G Punit Tunit
 | PHTint : forall (D:typeenv) (G:termenv) (n:nat),
     wfenv D G ->
     phastype D G (Pint n) Tint
 | PHTtrue : forall (D:typeenv) (G:termenv),
     wfenv D G ->
     phastype D G Ptrue Tbool
 | PHTfalse : forall (D:typeenv) (G:termenv),
     wfenv D G ->
     phastype D G Pfalse Tbool
 | PHTpair : forall (D:typeenv) (G:termenv) (p1 p2:prg) (ty1 ty2:typ),
     phastype D G p1 ty1 ->
     phastype D G p2 ty2 ->
     phastype D G (Ppair p1 p2) (Tprod ty1 ty2)
 | PHTfst : forall (D:typeenv) (G:termenv) (p:prg) (ty1 ty2:typ),
     phastype D G p (Tprod ty1 ty2) ->
     phastype D G (Pfst p) ty1
 | PHTsnd : forall (D:typeenv) (G:termenv) (p:prg) (ty2 ty1:typ),
     phastype D G p (Tprod ty1 ty2) ->
     phastype D G (Psnd p) ty2
 | PHTinl : forall (D:typeenv) (G:termenv) (ty2:typ) (p:prg) (ty1:typ),
     phastype D G p ty1 ->
     wftyp D G ty2 ->
     phastype D G (Pinl ty2 p) (Tsum ty1 ty2)
 | PHTinr : forall (D:typeenv) (G:termenv) (ty1:typ) (p:prg) (ty2:typ),
     phastype D G p ty2 ->
     wftyp D G ty1 ->
     phastype D G (Pinr ty1 p) (Tsum ty1 ty2)
 | PHTif : forall (D:typeenv) (G:termenv) (p0 p1 p2:prg) (ty:typ),
     phastype D G p0 Tbool ->
     phastype D G p1 ty ->
     phastype D G p2 ty ->
     phastype D G (Pif p0 p1 p2) ty
 | PHTcase : forall (D:typeenv) (G:termenv) (x:termvar) (p0 p1 p2:prg) (ty ty1 ty2:typ),
     phastype D G p0 (Tsum ty1 ty2) ->
     phastype D  (cons ( x , ty1 )  G )  p1 ty ->
     phastype D  (cons ( x , ty2 )  G )  p2 ty ->
     phastype D G (Pcase x p0 p1 p2) ty
 | PHTroll : forall (D:typeenv) (G:termenv) (ty:typ) (p:prg),
     phastype D G p  (vsubst 0  (Tmu ty)   ty )  ->
     wftyp D G (Tmu ty) ->
     phastype D G (Proll ty p) (Tmu ty)
 | PHTunroll : forall (D:typeenv) (G:termenv) (p:prg) (ty:typ),
     phastype D G p (Tmu ty) ->
     phastype D G (Punroll p)  (vsubst 0  (Tmu ty)   ty ) 
 | PHTfix : forall (D:typeenv) (G:termenv) (f x:termvar) (ty1 ty2:typ) (p:prg),
     phastype D  (cons ( x , ty1 )   (cons ( f , (Tarrow ty1 ty2) )  G )  )  p ty2 ->
     wftyp D G ty1 ->
     wftyp D G ty2 ->
     phastype D G (Pfix f x ty1 ty2 p) (Tarrow ty1 ty2)
 | PHTapp : forall (D:typeenv) (G:termenv) (p p':prg) (ty2 ty1:typ),
     phastype D G p (Tarrow ty1 ty2) ->
     phastype D G p' ty1 ->
     phastype D G (Papp p p') ty2
 | PHTgen : forall (D:typeenv) (G:termenv) (p:prg) (ty:typ) (aa:typename),
     phastype  (cons  aa   D )  G p  (vsubst 0  (Tname aa)   ty )  ->
     wftyp D G (Tforall ty) ->
      (~ In  aa   D )  ->
     phastype D G (Pgen p) (Tforall ty)
 | PHTinst : forall (D:typeenv) (G:termenv) (p:prg) (ty' ty:typ),
     phastype D G p (Tforall ty) ->
     wftyp D G ty' ->
     phastype D G (Pinst p ty')  (vsubst 0  ty'   ty ) 
 | PHTpack : forall (D:typeenv) (G:termenv) (ty':typ) (p:prg) (ty:typ),
     phastype D G p  (vsubst 0  ty'   ty )  ->
     wftyp D G ty' ->
     phastype D G (Ppack ty' p ty) (Texists ty)
 | PHTunpack : forall (D:typeenv) (G:termenv) (x:termvar) (p1 p2:prg) (ty2 ty:typ) (aa:typename),
     phastype D G p1 (Texists ty) ->
     phastype  (cons  aa   D )   (cons ( x ,  (vsubst 0  (Tname aa)   ty )  )  G )  p2 ty2 ->
     wftyp D G ty2 ->
      (~ In  aa   D )  ->
     phastype D G (Punpack x p1 p2) ty2
 | PHTnew : forall (D:typeenv) (G:termenv) (p:prg) (ty:typ),
     phastype D G p ty ->
     phastype D G (Pnew p) (Tref ty)
 | PHTread : forall (D:typeenv) (G:termenv) (p:prg) (ty:typ),
     phastype D G p (Tref ty) ->
     phastype D G (Pread p) ty
 | PHTwrite : forall (D:typeenv) (G:termenv) (p1 p2:prg) (ty:typ),
     phastype D G p1 (Tref ty) ->
     phastype D G p2 ty ->
     phastype D G (Pwrite p1 p2) Tunit
 | PHTrefeq : forall (D:typeenv) (G:termenv) (p1 p2:prg) (ty:typ),
     phastype D G p1 (Tref ty) ->
     phastype D G p2 (Tref ty) ->
     phastype D G (Prefeq p1 p2) Tbool.
(** definitions *)

(* defns CtxJudg *)
Inductive wfctx : ctx -> typeenv -> termenv -> typ -> typeenv -> termenv -> typ -> Prop :=    (* defn wfctx *)
 | wfctx_hole : forall (D:typeenv) (G:termenv) (ty:typ),
     wftyp D G ty ->
     wfctx Chole D G ty D G ty
 | wfctx_pair1 : forall (C:ctx) (p:prg) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty' ty2:typ),
     wfctx C D G ty D' G' ty' ->
     phastype D' G' p ty2 ->
     wfctx (Cpair1 C p) D G ty D' G' (Tprod ty' ty2)
 | wfctx_pair2 : forall (p:prg) (C:ctx) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty1 ty':typ),
     wfctx C D G ty D' G' ty' ->
     phastype D' G' p ty1 ->
     wfctx (Cpair2 p C) D G ty D' G' (Tprod ty1 ty')
 | wfctx_inl : forall (ty2:typ) (C:ctx) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty':typ),
     wfctx C D G ty D' G' ty' ->
     wftyp D' G' ty2 ->
     wfctx (Cinl ty2 C) D G ty D' G' (Tsum ty' ty2)
 | wfctx_inr : forall (ty1:typ) (C:ctx) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty':typ),
     wfctx C D G ty D' G' ty' ->
     wftyp D' G' ty1 ->
     wfctx (Cinr ty1 C) D G ty D' G' (Tsum ty1 ty')
 | wfctx_fst : forall (C:ctx) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty1 ty2:typ),
     wfctx C D G ty D' G' (Tprod ty1 ty2) ->
     wfctx (Cfst C) D G ty D' G' ty1
 | wfctx_snd : forall (C:ctx) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty2 ty1:typ),
     wfctx C D G ty D' G' (Tprod ty1 ty2) ->
     wfctx (Csnd C) D G ty D' G' ty2
 | wfctx_if0 : forall (C:ctx) (p1 p2:prg) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty':typ),
     wfctx C D G ty D' G' Tbool ->
     phastype D' G' p1 ty' ->
     phastype D' G' p2 ty' ->
     wfctx (Cif0 C p1 p2) D G ty D' G' ty'
 | wfctx_if1 : forall (p0:prg) (C:ctx) (p2:prg) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty':typ),
     wfctx C D G ty D' G' ty' ->
     phastype D' G' p0 Tbool ->
     phastype D' G' p2 ty' ->
     wfctx (Cif1 p0 C p2) D G ty D' G' ty'
 | wfctx_if2 : forall (p0 p1:prg) (C:ctx) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty':typ),
     wfctx C D G ty D' G' ty' ->
     phastype D' G' p0 Tbool ->
     phastype D' G' p1 ty' ->
     wfctx (Cif2 p0 p1 C) D G ty D' G' ty'
 | wfctx_case0 : forall (x:termvar) (C:ctx) (p1 p2:prg) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty' ty1 ty2:typ),
     wfctx C D G ty D' G' (Tsum ty1 ty2) ->
     phastype D'  (cons ( x , ty1 )  G' )  p1 ty' ->
     phastype D'  (cons ( x , ty2 )  G' )  p2 ty' ->
     wfctx (Ccase0 x C p1 p2) D G ty D' G' ty'
 | wfctx_case1 : forall (x:termvar) (p0:prg) (C:ctx) (p2:prg) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty' ty1 ty2:typ),
     phastype D' G' p0 (Tsum ty1 ty2) ->
     wfctx C D G ty D'  (cons ( x , ty1 )  G' )  ty' ->
     phastype D'  (cons ( x , ty2 )  G' )  p2 ty' ->
     wfctx (Ccase1 x p0 C p2) D G ty D' G' ty'
 | wfctx_case2 : forall (x:termvar) (p0 p1:prg) (C:ctx) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty' ty1 ty2:typ),
     phastype D' G' p0 (Tsum ty1 ty2) ->
     phastype D'  (cons ( x , ty1 )  G' )  p1 ty' ->
     wfctx C D G ty D'  (cons ( x , ty2 )  G' )  ty' ->
     wfctx (Ccase2 x p0 p1 C) D G ty D' G' ty'
 | wfctx_roll : forall (ty':typ) (C:ctx) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv),
     wftyp D' G' (Tmu ty') ->
     wfctx C D G ty D' G'  (vsubst 0  (Tmu ty')   ty' )  ->
     wfctx (Croll ty' C) D G ty D' G' (Tmu ty')
 | wfctx_unroll : forall (C:ctx) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty':typ),
     wfctx C D G ty D' G' (Tmu ty') ->
     wfctx (Cunroll C) D G ty D' G'  (vsubst 0  (Tmu ty')   ty' ) 
 | wfctx_fix : forall (f x:termvar) (ty1 ty2:typ) (C:ctx) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv),
     wfctx C D G ty D'  (cons ( x , ty1 )   (cons ( f , (Tarrow ty1 ty2) )  G' )  )  ty2 ->
     wftyp D' G' ty1 ->
     wftyp D' G' ty2 ->
     wfctx (Cfix f x ty1 ty2 C) D G ty D' G' (Tarrow ty1 ty2)
 | wfctx_app1 : forall (C:ctx) (p:prg) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty2 ty1:typ),
     wfctx C D G ty D' G' (Tarrow ty1 ty2) ->
     phastype D' G' p ty1 ->
     wfctx (Capp1 C p) D G ty D' G' ty2
 | wfctx_app2 : forall (p:prg) (C:ctx) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty2 ty1:typ),
     phastype D' G' p (Tarrow ty1 ty2) ->
     wfctx C D G ty D' G' ty1 ->
     wfctx (Capp2 p C) D G ty D' G' ty2
 | wfctx_gen : forall (C:ctx) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty':typ) (aa:typename),
     wfctx C D G ty  (cons  aa   D' )  G'  (vsubst 0  (Tname aa)   ty' )  ->
     wftyp D' G' (Tforall ty') ->
      (~ In  aa   D' )  ->
     wfctx (Cgen C) D G ty D' G' (Tforall ty')
 | wfctx_inst : forall (C:ctx) (ty1:typ) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty':typ),
     wfctx C D G ty D' G' (Tforall ty') ->
     wftyp D' G' ty1 ->
     wfctx (Cinst C ty1) D G ty D' G'  (vsubst 0  ty1   ty' ) 
 | wfctx_pack : forall (ty1:typ) (C:ctx) (ty':typ) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv),
     wfctx C D G ty D' G'  (vsubst 0  ty1   ty' )  ->
     wftyp D' G' ty1 ->
     wfctx (Cpack ty1 C ty') D G ty D' G' (Texists ty')
 | wfctx_unpack1 : forall (x:termvar) (C:ctx) (p:prg) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty2 ty':typ) (aa:typename),
     wfctx C D G ty D' G' (Texists ty') ->
     phastype  (cons  aa   D' )   (cons ( x ,  (vsubst 0  (Tname aa)   ty' )  )  G' )  p ty2 ->
     wftyp D' G' ty2 ->
      (~ In  aa   D' )  ->
     wfctx (Cunpack1 x C p) D G ty D' G' ty2
 | wfctx_unpack2 : forall (x:termvar) (p:prg) (C:ctx) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty2 ty':typ) (aa:typename),
     phastype D' G' p (Texists ty') ->
     wfctx C D G ty  (cons  aa   D' )   (cons ( x ,  (vsubst 0  (Tname aa)   ty' )  )  G' )  ty2 ->
     wftyp D' G' ty2 ->
      (~ In  aa   D' )  ->
     wfctx (Cunpack2 x p C) D G ty D' G' ty2
 | wfctx_new : forall (C:ctx) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty':typ),
     wfctx C D G ty D' G' ty' ->
     wfctx (Cnew C) D G ty D' G' (Tref ty')
 | wfctx_read : forall (C:ctx) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty':typ),
     wfctx C D G ty D' G' (Tref ty') ->
     wfctx (Cread C) D G ty D' G' ty'
 | wfctx_write1 : forall (C:ctx) (p:prg) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty':typ),
     wfctx C D G ty D' G' (Tref ty') ->
     phastype D' G' p ty' ->
     wfctx (Cwrite1 C p) D G ty D' G' Tunit
 | wfctx_write2 : forall (p:prg) (C:ctx) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty':typ),
     phastype D' G' p (Tref ty') ->
     wfctx C D G ty D' G' ty' ->
     wfctx (Cwrite2 p C) D G ty D' G' Tunit
 | wfctx_refeq1 : forall (C:ctx) (p:prg) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty':typ),
     wfctx C D G ty D' G' (Tref ty') ->
     phastype D' G' p (Tref ty') ->
     wfctx (Crefeq1 C p) D G ty D' G' Tbool
 | wfctx_refeq2 : forall (p:prg) (C:ctx) (D:typeenv) (G:termenv) (ty:typ) (D':typeenv) (G':termenv) (ty':typ),
     phastype D' G' p (Tref ty') ->
     wfctx C D G ty D' G' (Tref ty') ->
     wfctx (Crefeq2 p C) D G ty D' G' Tbool.

Lemma app_cont_of_list_scont : forall l e, 
  app_cont (cont_of_list_scont l) e = fold_right appctx_scont_exp e l.
Proof. induction l as [|[]?]; simpl; intros; try rewrite IHl; trivial. Qed. 

Lemma appctx_scont_of_cont : forall k e, 
  fold_right appctx_scont_exp e (list_scont_of_cont k) = app_cont k e.
Proof. induction k; simpl; intros; trivial; rewrite IHk; trivial. Qed. 

Lemma Rctx : 
  forall h e h' e' (RED: reduce h e h' e') k, 
     reduce h (app_cont k e) h' (app_cont k e').    
Proof.
  intros; rewrite <- (cont_of_list_scont_of_cont k), !app_cont_of_list_scont.
  induction (list_scont_of_cont k); simpl; auto using Rsctx.
Qed.

Hint Constructors hastype : hastype.
Hint Constructors phastype : hastype.

Lemma hastype_erase :
  forall D G p ty, phastype D G p ty -> hastype D G (erase_prg p) ty.
Proof. induction 1; simpl in *; eauto with hastype. Qed.

Lemma phastype_app_ctx :
  forall D G C ty D' G' ty', 
    wfctx C D G ty D' G' ty' -> forall p,
    phastype D G p ty -> 
    phastype D' G' (app_ctx C p) ty'.
Proof. induction 1; intros; simpl in *; eauto with hastype. Qed.




