Require Import Vtac Bool List Arith basis lang.
Require Import Coq.omega.Omega.
Require Import MinMax Vtac.
Set Implicit Arguments.


(*----------------------------------------------------------------------------*)
(** * Type substitutions for type variables *)
(*----------------------------------------------------------------------------*)

Lemma lc_mon :
  forall ty n m (LT: n <= m), lc_typ n ty -> lc_typ m ty.
Proof.
  induction ty; simpl in *; intros; intuition eauto using le_n_S; omega.
Qed.

Lemma lc_vsubst :
  forall tyA ty x, 
    lc_typ 0 tyA -> 
    lc_typ (S x) ty -> 
    lc_typ x (vsubst x tyA ty).
Proof.
  induction ty; simpl in *; intros; intuition eauto. 
  destruct (nat_compare_spec bb x); simpl; eauto using lc_mon, le_0_n.
  omega. 
Qed.

Lemma lc_vsubstD:
  forall tyA ty n,
    lc_typ n (vsubst n tyA ty) ->
    lc_typ (S n) ty.
Proof.
  induction ty; simpl; intros; f_equal; intuition auto.
  destruct (nat_compare_spec bb n); simpl in *; omega.
Qed.

Lemma vsubst_id : 
  forall tyA ty x, lc_typ x ty -> vsubst x tyA ty = ty.
Proof.
  induction ty; intross; des; f_equal; eauto.
  destruct (nat_compare_spec bb x); try (cut False; [|omega]); try done.
Qed.

Lemma in_ftn_vsubst_mon : 
  forall aa tyA ty n (IN: In aa (ftn_typ ty)), In aa (ftn_typ (vsubst n tyA ty)).
Proof.
  induction ty; intross; try eapply in_app_or in IN; des;
  try specialize (IHty1 _ IN); try specialize (IHty2 _ IN); des;
  eauto using in_or_app. 
Qed.

Lemma ftn_vsubst_mon_closed : 
  forall tyA ty n (CLO: ftn_typ (vsubst n tyA ty) = nil), ftn_typ ty = nil.
Proof.
  induction ty; intross; try eapply app_eq_nil in CLO; des;
  try specialize (IHty1 _ CLO); try specialize (IHty2 _ CLO0);
  try rewrite IHty1, IHty2; eauto.
Qed.

Lemma in_ftn_vsubst : 
  forall aa tyA ty n (IN: In aa (ftn_typ (vsubst n tyA ty))),
    In aa (ftn_typ tyA) \/ In aa (ftn_typ ty).
Proof.
  induction ty; intross; try eapply in_app_or in IN; des;
  try specialize (IHty1 _ IN); try specialize (IHty2 _ IN); des;
  eauto using in_or_app. 
  destruct nat_compare; vauto.
Qed.

Lemma ftn_vsubst_closed : 
  forall tyA ty n (CLO: ftn_typ tyA = nil) (CLO': ftn_typ ty = nil),
    ftn_typ (vsubst n tyA ty) = nil.
Proof.
  induction ty; intross; try eapply app_eq_nil in CLO'; des; 
  try specialize (IHty1 n CLO CLO'); try specialize (IHty2 n CLO CLO'0);
  try rewrite IHty1, IHty2; eauto.
  destruct nat_compare; vauto.
Qed.

(*----------------------------------------------------------------------------*)
(** * Type substitutions for free type names *)
(*----------------------------------------------------------------------------*)

Fixpoint tsubst_closed (sub : list (typename * typ)) :=
  match sub with
    | nil => True
    | h :: l => ftn_typ (snd h) = nil /\ tsubst_closed l
  end.

Lemma tsubst_closed_app : 
  forall l l',
    tsubst_closed (l ++ l') <-> tsubst_closed l /\ tsubst_closed l'.
Proof. intros; induction l; simpl; tauto. Qed.

Lemma tsubst_closed_minus2 : 
  forall l l',
    tsubst_closed l ->
    tsubst_closed (list_minus2 eq_typename l l').
Proof. induction l; simpl; intros; [|destruct list_mem]; simpl; intuition. Qed.
Hint Resolve tsubst_closed_minus2 : subst_lemmas.

Lemma tsubst_typ_disj : 
  forall x sub (H: list_disj eq_typename (map (@fst _ _) sub) (ftn_typ x)), tsubst_typ sub x = x.
Proof. induction x; subst_tac. Qed. 
Hint Resolve tsubst_typ_disj : subst_lemmas.

Lemma tsubst_typ_closed : 
  forall x (H: ftn_typ x = nil) sub, tsubst_typ sub x = x.
Proof. intros; apply tsubst_typ_disj; rewrite H; constructor. Qed.
Hint Resolve tsubst_typ_closed : subst_lemmas.
Hint Resolve (fun x H sub => eq_sym (tsubst_typ_closed x H sub)) : subst_lemmas. 

Lemma tsubst_typ_nil : forall x, tsubst_typ nil x = x.
Proof. induction x; subst_tac. Qed. 
Hint Resolve tsubst_typ_nil : subst_lemmas.
Hint Rewrite tsubst_typ_nil : subst_rw_lemmas.

Lemma tsubst_typ_cons : 
  forall x y z l (CLOSED: ftn_typ y = nil),
    tsubst_typ ((x,y)::l) z = tsubst_typ l (tsubst_typ (cons (x,y) nil) z). 
Proof. induction z; subst_tac'. Qed.
Hint Resolve tsubst_typ_cons : subst_lemmas.

Lemma tsubst_typ_remdup :
  forall x l, tsubst_typ (list_remdup2 eq_typename l) x = tsubst_typ l x.
Proof. induction x; simpl; intros; autorewrite with subst_rw_lemmas; f_equal; auto. Qed.
Hint Rewrite tsubst_typ_remdup : subst_rw_lemmas.

Lemma tsubst_typ_app :
  forall l (C: tsubst_closed l) l' x,
  tsubst_typ (l ++ l') x = tsubst_typ l' (tsubst_typ l x). 
Proof.
  induction l as [|[? ?] ? IH]; intros; simpl in *;
   [ autorewrite with subst_rw_lemmas
   | rewrite tsubst_typ_cons, IH, <- tsubst_typ_cons]; tauto.
Qed.

Lemma tsubst_typ_app_flip :
  forall l l' (D: list_disj eq_typename (map (@fst _ _) l) (map (@fst _ _) l')) x,
  tsubst_typ (l ++ l') x = tsubst_typ (l' ++ l) x. 
Proof.
  intros; revert l l' D; induction x; intros; simpl; 
    rewrite ?list_minus2_app, ?(list_assoc_flip _ _ _ D); subst_tac'.
Qed.

Lemma tsubst_typ_app_disj :
  forall l' (C: tsubst_closed l') l
   (D: list_disj eq_typename (map (@fst _ _) l) (map (@fst _ _) l')) e,
  tsubst_typ l (tsubst_typ l' e) = tsubst_typ (l ++ l') e.
Proof.
  intros; rewrite <- tsubst_typ_app, <- tsubst_typ_app_flip; auto.
Qed.

Lemma tsubst_typ_comm :
  forall l (C: tsubst_closed l) l' (C': tsubst_closed l')
   (D: list_disj eq_typename (map (@fst _ _) l) (map (@fst _ _) l')) e,
    tsubst_typ l (tsubst_typ l' e) = (tsubst_typ l' (tsubst_typ l e)).
Proof.
  intros; rewrite <- !tsubst_typ_app, (tsubst_typ_app_flip _ _ D); auto.
Qed.

Lemma ftn_tsubst_typ :
  forall l (C: tsubst_closed l) x,
  ftn_typ (tsubst_typ l x) = list_minus eq_typename (ftn_typ x) (map (@fst _ _) l).
Proof.
  intros; revert l C; induction x; simpl; intros; trivial; rewrite ?list_minus_app;
  repeat match goal with 
           | |- _ => solve [auto]
           | |- _ ++ _ = _ ++ _ => f_equal
           | H : _ |- _ => solve[rewrite H, list_map_fst_minus2, list_minus_comm, 
                                         list_minus_red; auto with subst_lemmas] end.
  induction l as [|[]]; simpl in *; [|destruct (eq_typename)]; intuition. 
Qed.

Lemma tsubst_typ_cons_minus2: 
  forall x e' l e,
    tsubst_typ ((x, e') :: list_minus2 eq_typename l (x :: nil)) e
    = tsubst_typ ((x, e') :: l) e.
Proof.
  intros.
  rewrite <- (tsubst_typ_remdup _ (_ :: list_minus2 _ _ _)).
  rewrite <- (tsubst_typ_remdup _ (_ :: l)); simpl.
  rewrite <- list_minus2_remdup2, <- list_minus2_app_r.
  rewrite list_minus2_eq with (l2 := x :: nil); simpl; tauto.
Qed.

Lemma tsubst_typ_eq:
  forall l l' e,
    list_remdup2 eq_typename l = list_remdup2 eq_typename l' ->
    tsubst_typ l e = tsubst_typ l' e.
Proof. intros; rewrite <- tsubst_typ_remdup, H, tsubst_typ_remdup; auto. Qed.

Lemma lc_tsubst:
  forall d (CLOSd : forall x, In x d -> lc_typ 0 (snd x)) 
    ty n (H: lc_typ n ty),
    lc_typ n (tsubst_typ d ty).
Proof.
  induction ty; intross; des; auto.
  pose proof (list_assoc_spec eq_typename aa d).
  destruct list_assoc; eauto.
  apply CLOSd in H0; simpl in *; eauto using lc_mon, le_0_n.
Qed.

Lemma tsubst_lc: forall s ty n (LC: lc_typ n (tsubst_typ s ty)), lc_typ n ty.
Proof. by intros s ty; induction ty; intros; simpls; des; eauto. Qed.

Lemma tsubst_typ_vsubst : 
  forall d tyA ty x
    (IN: forall x, In x d -> lc_typ 0 (snd x)),
    lc_typ (S x) ty ->
    tsubst_typ d (vsubst x tyA ty) = vsubst x (tsubst_typ d tyA) (tsubst_typ d ty).
Proof.
  induction ty; intross; des; f_equal; eauto.
  - pose proof (list_assoc_spec eq_typename aa d);
    destruct list_assoc; simpls.
    by apply IN in H0; simpl in *; rewrite vsubst_id; eauto using lc_mon, le_0_n.
  - by destruct nat_compare; simpls.
Qed.

Lemma tsubst_typ_nin:
   forall aa tyA d ty, 
     ~ In aa (ftn_typ ty) ->
   tsubst_typ ((aa, tyA) :: d) ty = tsubst_typ d ty.
Proof.
    induction ty; intross; f_equal; eauto;
    repeat match goal with
      | H : ~ In _ _ |- ~ In _ _ => intro; destruct H
      | |- In _ (_ ++ _) => apply in_or_app; auto
      | |- In _ (list_minus _ _ _) => apply in_list_minusI; simpl; auto
      | H: _ |- _ => apply H 
      | |- context[if (eq_typename ?x ?y) then _ else _] => destruct (eq_typename x y); subst; try done
    end; tauto.
Qed.

(*----------------------------------------------------------------------------*)
(** * Exp substitutions *)
(*----------------------------------------------------------------------------*)

Fixpoint subst_closed (sub : list (termvar * exp)) :=
  match sub with
    | nil => True
    | h :: l => fv_exp (snd h) = nil /\ subst_closed l
  end.

Lemma subst_closed_app : 
  forall l l',
    subst_closed (l ++ l') <-> subst_closed l /\ subst_closed l'.
Proof. intros; induction l; simpl; tauto. Qed.

Lemma subst_closed_minus2 : 
  forall l l',
    subst_closed l ->
    subst_closed (list_minus2 eq_termvar l l').
Proof. induction l; simpl; intros; [|destruct list_mem]; simpl; intuition. Qed.
Hint Resolve subst_closed_minus2 : subst_lemmas.

Lemma subst_exp_disj : 
  forall x sub (H: list_disj eq_termvar (map (@fst _ _) sub) (fv_exp x)), subst_exp sub x = x.
Proof. induction x; subst_tac. Qed. 
Hint Resolve subst_exp_disj : subst_lemmas.

Lemma subst_val_disj : 
  forall x sub (H: list_disj eq_termvar (map (@fst _ _) sub) (fv_val x)), subst_val sub x = x.
Proof. induction x; subst_tac. Qed.
Hint Resolve subst_val_disj : subst_lemmas.

Lemma subst_cont_disj : 
  forall x sub (H: list_disj eq_termvar (map (@fst _ _) sub) (fv_cont x)), subst_cont sub x = x.
Proof. induction x; subst_tac. Qed.
Hint Resolve subst_cont_disj : subst_lemmas.

Lemma subst_exp_closed : 
  forall x (H: fv_exp x = nil) sub, subst_exp sub x = x.
Proof. intros; apply subst_exp_disj; rewrite H; constructor. Qed.
Hint Resolve subst_exp_closed : subst_lemmas.
Hint Resolve (fun x H sub => eq_sym (subst_exp_closed x H sub)) : subst_lemmas. 

Lemma subst_val_closed : 
  forall x (H: fv_val x = nil) sub, subst_val sub x = x.
Proof. intros; apply subst_val_disj; rewrite H; constructor. Qed.
Hint Resolve subst_exp_closed : subst_lemmas.
Hint Resolve (fun x H sub => eq_sym (subst_val_closed x H sub)) : subst_lemmas. 

Lemma subst_cont_closed : 
  forall x (H: fv_cont x = nil) sub, subst_cont sub x = x.
Proof. intros; apply subst_cont_disj; rewrite H; constructor. Qed.
Hint Resolve subst_exp_closed : subst_lemmas.
Hint Resolve (fun x H sub => eq_sym (subst_cont_closed x H sub)) : subst_lemmas. 

Lemma subst_exp_nil : forall x, subst_exp nil x = x.
Proof. induction x; subst_tac. Qed. 
Hint Resolve subst_exp_nil : subst_lemmas.
Hint Rewrite subst_exp_nil : subst_rw_lemmas.

Lemma subst_val_nil : forall x, subst_val nil x = x.
Proof. induction x; subst_tac. Qed. 
Hint Resolve subst_val_nil : subst_lemmas.
Hint Rewrite subst_val_nil : subst_rw_lemmas.

Lemma subst_cont_nil : forall x, subst_cont nil x = x.
Proof. induction x; subst_tac. Qed. 
Hint Resolve subst_cont_nil : subst_lemmas.
Hint Rewrite subst_cont_nil : subst_rw_lemmas.

Lemma subst_exp_cons : 
  forall x e' e l (CLOSED: fv_exp e' = nil),
    subst_exp ((x,e')::l) e = subst_exp l (subst_exp (cons (x,e') nil) e). 
Proof. induction e; subst_tac'. Qed.
Hint Resolve subst_exp_cons : subst_lemmas.

Lemma subst_val_cons : 
  forall x e' v l (CLOSED: fv_exp e' = nil),
    subst_val ((x,e')::l) v = subst_val l (subst_val (cons (x,e') nil) v). 
Proof. induction v; subst_tac'. Qed.
Hint Resolve subst_val_cons : subst_lemmas.

Lemma subst_cont_cons : 
  forall x e' k l (CLOSED: fv_exp e' = nil),
    subst_cont ((x,e')::l) k = subst_cont l (subst_cont (cons (x,e') nil) k). 
Proof. induction k; subst_tac'. Qed.
Hint Resolve subst_cont_cons : subst_lemmas.

Lemma subst_exp_remdup :
  forall x l, subst_exp (list_remdup2 eq_termvar l) x = subst_exp l x.
Proof. induction x; simpl; intros; autorewrite with subst_rw_lemmas; f_equal; auto. Qed.
Hint Rewrite subst_exp_remdup : subst_rw_lemmas.

Lemma subst_val_remdup :
  forall x l, subst_val (list_remdup2 eq_termvar l) x = subst_val l x.
Proof. induction x; simpl; intros; autorewrite with subst_rw_lemmas; f_equal; auto. Qed.
Hint Rewrite subst_val_remdup : subst_rw_lemmas.

Lemma subst_cont_remdup :
  forall x l, subst_cont (list_remdup2 eq_termvar l) x = subst_cont l x.
Proof. induction x; simpl; intros; autorewrite with subst_rw_lemmas; f_equal; auto. Qed.
Hint Rewrite subst_cont_remdup : subst_rw_lemmas.

Lemma subst_exp_app :
  forall l (C: subst_closed l) l' e,
  subst_exp (l ++ l') e = subst_exp l' (subst_exp l e). 
Proof.
  induction l as [|[? ?] ? IH]; intros; simpl in *;
   [ autorewrite with subst_rw_lemmas
   | rewrite subst_exp_cons, IH, <- subst_exp_cons]; tauto.
Qed.

Lemma subst_exp_app_flip :
  forall l l' (D: list_disj eq_termvar (map (@fst _ _) l) (map (@fst _ _) l')) e,
  subst_exp (l ++ l') e = subst_exp (l' ++ l) e. 
Proof.
  intros; revert l l' D; induction e; intros; simpl; 
    rewrite ?list_minus2_app, ?(list_assoc_flip _ _ _ D); subst_tac'.
Qed.

Lemma subst_exp_app_disj :
  forall l' (C: subst_closed l') l
   (D: list_disj eq_termvar (map (@fst _ _) l) (map (@fst _ _) l')) e,
  subst_exp l (subst_exp l' e) = subst_exp (l ++ l') e.
Proof.
  intros; rewrite <- subst_exp_app, <- subst_exp_app_flip; auto.
Qed.

Lemma subst_exp_comm :
  forall l (C: subst_closed l) l' (C': subst_closed l')
   (D: list_disj eq_termvar (map (@fst _ _) l) (map (@fst _ _) l')) e,
    subst_exp l (subst_exp l' e) = (subst_exp l' (subst_exp l e)).
Proof.
  intros; rewrite <- !subst_exp_app, (subst_exp_app_flip _ _ D); auto.
Qed.


Lemma fv_subst_exp :
  forall l (C: subst_closed l) e,
  fv_exp (subst_exp l e) = list_minus eq_termvar (fv_exp e) (map (@fst _ _) l).
Proof.
  intros; revert l C; induction e; simpl; intros; trivial; rewrite ?list_minus_app;
  repeat match goal with 
           | |- _ => solve [auto]
           | |- _ ++ _ = _ ++ _ => f_equal
           | H : _ |- _ => solve[rewrite H, list_map_fst_minus2, list_minus_comm, 
                                         list_minus_red; auto with subst_lemmas] end.
  induction l as [|[]]; simpl in *; [|destruct (eq_termvar)]; intuition. 
Qed.

Lemma subst_exp_cons_minus2: 
  forall x e' l e,
    subst_exp ((x, e') :: list_minus2 eq_termvar l (x :: nil)) e
    = subst_exp ((x, e') :: l) e.
Proof.
  intros.
  rewrite <- (subst_exp_remdup _ (_ :: list_minus2 _ _ _)).
  rewrite <- (subst_exp_remdup _ (_ :: l)); simpl.
  rewrite <- list_minus2_remdup2, <- list_minus2_app_r.
  rewrite list_minus2_eq with (l2 := x :: nil); simpl; tauto.
Qed.

Lemma subst_exp_eq:
  forall l l' e,
    list_remdup2 eq_termvar l = list_remdup2 eq_termvar l' ->
    subst_exp l e = subst_exp l' e.
Proof. intros; rewrite <- subst_exp_remdup, H, subst_exp_remdup; auto. Qed.
